/****************************************
*  Computer Algebra System SINGULAR     *
****************************************/
/* $Id: mmallocb.c,v 1.17 1998/06/13 14:11:51 Singular Exp $ */

/*
* ABSTRACT:
*/

#define _POSIX_SOURCE 1

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "omalloc.h"
#include "mmprivat.h"

static char * mm_theHeapPtr = NULL;
static char * mm_tmpHeapPtr = NULL;
static char * mm_normHeapPtr = NULL;

static int mm_theBytesLeft = 0;
static int mm_tmpBytesLeft = 0;
static int mm_normBytesLeft = 0;

static status_t mm_heapStatus = MM_NORMAL;

static void mmDistributeRestOfHeap( void );

void* mmGetMemory( size_t size )
{
  void * dummy;

  if ( mm_theBytesLeft < size )
  {
    mmDistributeRestOfHeap();
    mm_theHeapPtr = (char*)mmGetBlock();
    mm_theBytesLeft = MAXDATA;
  }
  dummy = (void*)mm_theHeapPtr;
  mm_theBytesLeft -= size;
  mm_theHeapPtr += size;
  return dummy;
}

void * mmAllocBlock( size_t size )
{
  int i;

  if (size == 0)
    printf("SIZE == 0\n");

/* assume(size > 0); */

  i = mmGetIndex( size );

  if ( i < 0 )
  {
    void *res=(void*)malloc( size );
    if (res==NULL)
    {
      printf("out of memory\n");
      exit(99);
    }
    mm_bytesReal += size;
    return res;
  }
  else
  {
    mcb result;
    size = mmGetSize( i );
    if ( ( result = mm_theList[i] ) == NULL )
    {
      return (void*)mmGetMemory( size );
    }
    else
    {
      mm_theList[i] = (mcb)*result;
      return result;
    }
  }
}

void * mmAllocBlock0( size_t size )
{
  void *result=mmAllocBlock(size);
  memset(result,0,size);
  return result;
}

void mmFreeBlock( void* adr, size_t size )
{

  if ( ( adr == NULL ) || ( size == 0 ) )
    return;

  {
    int i = mmGetIndex( size );

    if ( i < 0 )
    {
      mm_bytesReal -= size;
      free( adr );
    }
    else
    {
      size = mmGetSize( i );
      {
        /* This used to be mmPutMemory -- let's "inline" it manually */
        void * dummy=(void *)((char *)adr+size);

        if ( dummy == mm_theHeapPtr)
        {
          mm_theHeapPtr=(char *)adr;
          mm_theBytesLeft += size;
          return;
        }
      }
      *(mcb)adr = mm_theList[i];
      mm_theList[i] = (mcb)adr;
    }
  }
}

void * mmReallocBlock( void* adr, size_t oldsize, size_t newsize )
{
  int i = (oldsize == 0 ? -1: mmGetIndex( oldsize ));
  /*  assume(newsize > 0); */
  int j = mmGetIndex( newsize );

  if ( ( i < 0 ) && ( j < 0 ) )
  {
    void *res=(void *)realloc( adr, newsize );
    if (res==NULL)
    {
      printf("out of memory\n");
      exit(99);
    }
    mm_bytesReal += ( (int)newsize - (int)oldsize );
    return res;
  }
  else if ( i == j )
    return adr;
  else
  {
    /* since we know i and j this can be done better, this is the quick way */
    void *newadr = mmAllocBlock( newsize );
    memcpy( newadr, adr, (oldsize < newsize) ? oldsize : newsize );
    mmFreeBlock( adr, oldsize );
    return newadr;
  }
}

static void mmDistributeRestOfHeap( void )
{

  char* dummy;
  int j;

  while ( mm_theBytesLeft > RealSizeFromSize( mmGetSize( 0 ) ) )
  {
    j = mmGetIndex( SizeFromRealSize( mm_theBytesLeft ) );
    if ( RealSizeFromSize( mmGetSize( j ) ) > mm_theBytesLeft ) j--;
    dummy = mm_theHeapPtr;
    mm_theHeapPtr += RealSizeFromSize( mmGetSize( j ) );
    mm_theBytesLeft -= RealSizeFromSize( mmGetSize( j ) );
    *(mcb)dummy = mm_theList[j];
    mm_theList[j] = (mcb)dummy;
  }
}
